"""Exec module for managing subnets."""
from collections import OrderedDict
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(hub, ctx, resource_id: str, raw: bool = False) -> Dict[str, Any]:
    """Get subnet resource from resource_id.

    Args:
        resource_id(str):
            The resource_id of subnet
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.network.subnets.get resource_id="value"  raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.subnets.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}/subnets/{subnet_name}"
                    raw: False
    """
    result = dict(comment=[], result=True, ret=None)
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "virtualNetworks": "virtual_network_name",
            "subnets": "subnet_name",
        }
    )

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{hub.exec.azure.URL}{resource_id}?api-version=2021-06-01",
        success_codes=[200, 201],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"] = response_get["comment"]
        return result

    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )
    if raw:
        result["ret"] = response_get["ret"]

    else:
        result["ret"] = hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
            resource=response_get["ret"],
            idem_resource_name=resource_id,
            resource_id=resource_id,
            **uri_parameter_values,
        )

    return result


async def list_(hub, ctx) -> Dict:
    """List of subnets

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.network.subnets.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.network.subnets.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "virtualNetworks": "virtual_network_name",
            "subnets": "subnet_name",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Network/virtualNetworks?api-version=2021-03-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for vnet in resource_list:
                if vnet.get("properties") and vnet.get("properties").get("subnets"):
                    subnet_list = vnet["properties"]["subnets"]
                    for resource in subnet_list:
                        resource_id = resource["id"]
                        uri_parameter_values = (
                            hub.tool.azure.uri.get_parameter_value_in_dict(
                                resource_id, uri_parameters
                            )
                        )
                        result["ret"].append(
                            hub.tool.azure.network.subnets.convert_raw_subnets_to_present(
                                resource=resource,
                                idem_resource_name=resource_id,
                                subscription_id=subscription_id,
                                resource_id=resource_id,
                                **uri_parameter_values,
                            )
                        )

    return result



"""States module for managing Firewall."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
__reconcile_wait__ = {"static": {"wait_in_seconds": 20}}


async def present(
    hub,
    ctx,
    name: str,
    location: str,
    resource_group_name: str,
    firewall_name: str,
    subscription_id: str = None,
    tags: Dict = None,
    resource_id: str = None,
    zones: List[str] = None,
    sku: make_dataclass(
        "sku",
        [("name", str, field(default=None)), ("tier", str, field(default=None))],
    ) = None,
    firewall_policy_id: str = None,
    ip_configuration: List[
        make_dataclass(
            "ipConfiguration",
            [
                ("name", str, field(default=None)),
                ("subnet_id", str, field(default=None)),
                ("public_ip_address_id", str, field(default=None)),
            ],
        )
    ] = None,
    management_ip_configuration: make_dataclass(
        "managementIpConfiguration",
        [
            ("name", str, field(default=None)),
            ("subnet_id", str, field(default=None)),
            ("public_ip_address_id", str, field(default=None)),
        ],
    ) = None,
) -> Dict:
    r"""Create or update firewall.

    Args:
        name(str): The identifier for this state.
        location(str): Resource location. Changing this forces a new resource to be created.
        resource_group_name(str): The name of the resource group.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): firewall resource id on Azure
        tags(dict[str, str], Optional): Resource tags.
        firewall_name(str): The name of the firewall.
        zones(list[str]): A list of availability zones denoting where the resource needs to come from.
        sku(dict[str, Any], Optional): The SKU of the Firewall.

            * name(str):
                SKU name of the Firewall. Possible values are AZFW_Hub and AZFW_VNet. Changing this forces a new resource to be created.
            * tier(str):
                SKU tier of the Firewall. Possible values are Premium, Standard and Basic.
        firewall_policy_id(str): The ID of the Firewall Policy applied to this Firewall.
        ip_configuration(list[dict[str, Any]], Optional): IP configuration of the Firewall.

            * name(str):
                SKU name of the Firewall. Possible values are AZFW_Hub and AZFW_VNet. Changing this forces a new resource to be created.
            * subnet_id(str):
                SKU tier of the Firewall. Possible values are Premium, Standard and Basic.
            * public_ip_address_id(str):
                SKU tier of the Firewall. Possible values are Premium, Standard and Basic.
        management_ip_configuration(dict[str, Any], Optional): Management IP configuration of the Firewall.

            * name(str):
                Specifies the name of the IP Configuration.
            * subnet_id(str):
                Reference to the subnet associated with the IP Configuration. Changing this forces a new resource to be created.
            * public_ip_address_id(str):
                The ID of the Public IP Address associated with the firewall.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.network.firewall.present:
                - name: my_firewall
                - subscription_id: my_sub_id
                - resource_group_name: my_rg-1
                - firewall_name: my-firewall
                - location: eastus
                - tags:
                    key: valuer
                - zones:
                    - 1
                - sku:
                    name: AZFW_VNet
                    tier: Premium
                - firewall_policy_id: my_fp_id
                - ip_configuration:
                    - name: name
                      subnet_id: my_sub_id
                      public_ip_address_id: my_public_ip_address
                - management_ip_configuration:
                      name: name
                      subnet_id: my_AzureFirewallManagementSubnet_subnet
                      public_ip_address_id: my_management_public_ip_addess

    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/azureFirewalls/{firewall_name}"
    response_get = await hub.exec.azure.network.firewall.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "subscription_id": subscription_id,
                        "firewall_name": firewall_name,
                        "tags": tags,
                        "location": location,
                        "resource_id": resource_id,
                        "zones": zones,
                        "sku": sku,
                        "firewall_policy_id": firewall_policy_id,
                        "ip_configuration": ip_configuration,
                        "management_ip_configuration": management_ip_configuration,
                    },
                )
                result["comment"].append(
                    f"Would create azure.network.firewall '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = (
                    hub.tool.azure.network.firewall.convert_present_to_raw_firewall(
                        location=location,
                        tags=tags,
                        zones=zones,
                        sku=sku,
                        firewall_policy_id=firewall_policy_id,
                        ip_configuration=ip_configuration,
                        management_ip_configuration=management_ip_configuration,
                    )
                )

                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                    success_codes=[200, 201],
                    json=payload,
                )
                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create firewall {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    firewall_name=firewall_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(f"Created azure.network.firewall '{name}'")
                return result

        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                firewall_name=firewall_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.network.firewall.update_firewall_payload(
                existing_resource, {"tags": tags}
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.network.firewall '{name}' doesn't need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        firewall_name=firewall_name,
                        resource_id=resource_id,
                        subscription_id=subscription_id,
                    )
                    result["comment"].append(
                        f"Would update azure.network.firewall '{name}'"
                    )
                return result

            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.firewall '{name}' doesn't need to be updated."
                )
                return result
            result["comment"] = result["comment"] + new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
                success_codes=[200, 201],
                json=new_payload["ret"],
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.network.firewall {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                firewall_name=firewall_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(f"Updated azure.network.firewall '{name}'")
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.firewall {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    firewall_name: str,
    subscription_id: str = None,
) -> dict:
    r"""Delete a firewall.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        firewall_name(str): The name of the firewall.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.firewall.absent:
                - name: my-fp
                - subscription_id: my-subscription
                - resource_group_name: my-resource-group
                - firewall_name: my-fp
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/azureFirewalls/{firewall_name}"
    response_get = await hub.exec.azure.network.firewall.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"]:
            result[
                "old_state"
            ] = hub.tool.azure.network.firewall.convert_raw_firewall_to_present(
                resource=response_get["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                firewall_name=firewall_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.network.firewall '{firewall_name}'"
                )
                return result
            if (
                response_get["ret"]["properties"]
                and response_get["ret"]["properties"]["provisioningState"] != "Deleting"
            ):
                response_delete = await hub.exec.request.raw.delete(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
                    success_codes=[200, 202],
                )

                if not response_delete["result"]:
                    hub.log.debug(
                        f"Could not delete azure.network.firewall {response_delete['comment']} {response_delete['ret']}"
                    )
                    result["result"] = False
                    result["comment"] = [
                        response_delete["comment"],
                        response_delete["ret"],
                    ]
                    return result

            result["comment"].append(
                f"Deleted azure.network.firewall '{firewall_name}'"
            )
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.network.firewall '{firewall_name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not azure.network.firewall '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all firewall under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.firewall
    """
    result = {}
    ret_list = await hub.exec.azure.network.firewall.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe network firewall {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.firewall.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



"""State module for managing Public IP Address."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    public_ip_address_name: str,
    location: str,
    allocation_method: str,
    zones: List[str] = None,
    ddos_protection_mode: str = None,
    ddos_protection_plan_id: str = None,
    domain_name_label: str = None,
    edge_zone: str = None,
    idle_timeout_in_minutes: int = None,
    ip_tags: Dict[str, str] = None,
    ip_version: str = None,
    public_ip_prefix_id: str = None,
    reverse_fqdn: str = None,
    sku: str = None,
    sku_tier: str = None,
    tags: Dict[str, str] = None,
    subscription_id: str = None,
    resource_id: str = None,
) -> Dict:
    r"""Create or update Public IP Addresses.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        public_ip_address_name(str): The name of the public IP address.
        location(str): Resource location.
        allocation_method(str): Defines the allocation method for this IP address.
        zones(list[str], Optional): A collection containing the availability zone to allocate the Public IP in.
        ddos_protection_mode(str, Optional): The DDoS protection mode of the public IP.
        ddos_protection_plan_id(str, Optional): The ID of DDoS protection plan associated with the public IP.
        domain_name_label(str, Optional): Label for the Domain Name.
        edge_zone(str, Optional): Specifies the Edge Zone within the Azure Region where this Public IP should exist.
        idle_timeout_in_minutes(int, Optional): Specifies the timeout for the TCP idle connection.
        ip_tags(dict, Optional): A mapping of IP tags to assign to the public IP.
        ip_version(str, Optional): The IP Version to use.
        public_ip_prefix_id(str, Optional): If specified then public IP address allocated will be provided from the public IP prefix resource.
        reverse_fqdn(str, Optional): A fully qualified domain name that resolves to this public IP address.
        sku(str, Optional): The SKU of the Public IP.
        sku_tier(str, Optional): The SKU Tier that should be used for the Public IP.
        tags(dict, Optional): Resource tags.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Management group resource id on Azure

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.network.public_ip_addresses.present:
                - name: value
                - resource_group_name: value
                - public_ip_address_name: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id

    if resource_id is None:
        resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}"

    response_get = await hub.exec.azure.network.public_ip_addresses.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if not response_get["ret"]:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "public_ip_address_name": public_ip_address_name,
                        "location": location,
                        "allocation_method": allocation_method,
                        "zones": zones,
                        "ddos_protection_mode": ddos_protection_mode,
                        "ddos_protection_plan_id": ddos_protection_plan_id,
                        "domain_name_label": domain_name_label,
                        "edge_zone": edge_zone,
                        "idle_timeout_in_minutes": idle_timeout_in_minutes,
                        "ip_tags": ip_tags,
                        "ip_version": ip_version,
                        "public_ip_prefix_id": public_ip_prefix_id,
                        "reverse_fqdn": reverse_fqdn,
                        "sku": sku,
                        "sku_tier": sku_tier,
                        "tags": tags,
                        "resource_id": resource_id,
                        "subscription_id": subscription_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.network.public_ip_addresses '{name}'"
                )
                return result
            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.network.public_ip_addresses.convert_present_to_raw_public_ip_addresses(
                    location=location,
                    allocation_method=allocation_method,
                    zones=zones,
                    ddos_protection_mode=ddos_protection_mode,
                    ddos_protection_plan_id=ddos_protection_plan_id,
                    domain_name_label=domain_name_label,
                    edge_zone=edge_zone,
                    idle_timeout_in_minutes=idle_timeout_in_minutes,
                    ip_tags=ip_tags,
                    ip_version=ip_version,
                    public_ip_prefix_id=public_ip_prefix_id,
                    reverse_fqdn=reverse_fqdn,
                    sku=sku,
                    sku_tier=sku_tier,
                    tags=tags,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
                    success_codes=[200, 201],
                    json=payload,
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create azure.network.public_ip_addresses {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"] = [response_put["comment"], response_put["ret"]]
                    result["result"] = False
                    return result

                result[
                    "new_state"
                ] = hub.tool.azure.network.public_ip_addresses.convert_raw_public_ip_addresses_to_present(
                    resource=response_put["ret"],
                    idem_resource_name=name,
                    resource_group_name=resource_group_name,
                    public_ip_address_name=public_ip_address_name,
                    resource_id=resource_id,
                    subscription_id=subscription_id,
                )
                result["comment"].append(
                    f"Created azure.network.public_ip_addresses '{name}'"
                )
                return result

        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.network.public_ip_addresses.convert_raw_public_ip_addresses_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                public_ip_address_name=public_ip_address_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.network.public_ip_addresses.update_public_ip_addresses_payload(
                existing_resource,
                {
                    "allocation_method": allocation_method,
                    "zones": zones,
                    "ddos_protection_mode": ddos_protection_mode,
                    "ddos_protection_plan_id": ddos_protection_plan_id,
                    "domain_name_label": domain_name_label,
                    "edge_zone": edge_zone,
                    "idle_timeout_in_minutes": idle_timeout_in_minutes,
                    "ip_tags": ip_tags,
                    "ip_version": ip_version,
                    "public_ip_prefix_id": public_ip_prefix_id,
                    "reverse_fqdn": reverse_fqdn,
                    "sku": sku,
                    "sku_tier": sku_tier,
                    "tags": tags,
                },
            )

            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.network.public_ip_addresses '{name}' has no property need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.network.public_ip_addresses.convert_raw_public_ip_addresses_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        public_ip_address_name=public_ip_address_name,
                        resource_id=resource_id,
                        subscription_id=subscription_id,
                    )
                    result["comment"].append(
                        f"Would update azure.network.public_ip_addresses '{name}'"
                    )
                return result

            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.network.public_ip_addresses '{name}' has no property need to be updated."
                )
                return result
            result["comment"] = result["comment"] + new_payload["comment"]
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
                success_codes=[200],
                json=new_payload["ret"],
            )
            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.network.public_ip_addresses {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_put["comment"], response_put["ret"]]
                return result

            result[
                "new_state"
            ] = hub.tool.azure.network.public_ip_addresses.convert_raw_public_ip_addresses_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                public_ip_address_name=public_ip_address_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            if result["old_state"] == result["new_state"]:
                result["comment"].append(
                    f"azure.network.public_ip_addresses '{name}' has no property need to be updated."
                )
                return result

            result["comment"].append(
                f"Updated azure.network.public_ip_addresses '{name}'"
            )
            return result

    else:
        hub.log.debug(
            f"Could not get azure.network.public_ip_addresses {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
        return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    public_ip_address_name: str,
    subscription_id: str = None,
) -> dict:
    r"""Delete Public IP Addresses.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        public_ip_address_name(str): The name of the public IP address.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.public_ip_addresses.absent:
                - name: value
                - resource_group_name: value
                - public_ip_address_name: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }

    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}"
    response_get = await hub.exec.azure.network.public_ip_addresses.get(
        ctx,
        resource_id=resource_id,
    )
    if response_get["result"]:
        if response_get["ret"]:
            result["old_state"] = response_get["ret"]
            result["old_state"]["name"] = name

            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.network.public_ip_addresses '{name}'"
                )
                return result
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
                success_codes=[200, 202, 204],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.network.public_ip_addresses '{name}' {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"] = [response_delete["comment"], response_delete["ret"]]
                return result

            result["comment"].append(
                f"Deleted azure.network.public_ip_addresses '{name}'"
            )
            return result

        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.network.public_ip_addresses '{name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.network.public_ip_addresses '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"] = [response_get["comment"], response_get["ret"]]
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Public IP Addresses under the same subscription.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure_auto.network.public_ip_addresses
    """
    result = {}
    ret_list = await hub.exec.azure.network.public_ip_addresses.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe public ip addresses {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.public_ip_addresses.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }
    return result



CLI_CONFIG = {
    "rest_api": {
        "subcommands": ["azure"],
        "dyne": "pop_create",
    },
    "resources": {
        "subcommands": ["azure"],
        "dyne": "pop_create",
    },
}
CONFIG = {
    "resources": {
        "default": [],
        "nargs": "*",
        "help": "The resources to target under the rest api, defaults to all",
        "dyne": "pop_create",
    },
    "rest_api": {
        "default": "Azure",
        "help": "The Microsoft REST API to target, defaults to Azure",
        "dyne": "pop_create",
    },
}
DYNE = {
    "pop_create": ["pop_create"],
    "tool": ["tool"],
    "states": ["states"],
    "exec": ["exec"],
    "acct": ["acct"],
    "reconcile": ["reconcile"],
}
SUBCOMMANDS = {
    "azure": {
        "help": "Create idem_azure state modules by Azure REST API web",
        "dyne": "pop_create",
    },
}



"""Exec module for managing Authorization Role Assignments."""
from collections import OrderedDict
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
    raw: bool = False,
) -> Dict[str, Any]:
    """Get role assignment from resource_id.

    Args:
        resource_id(str):
            The resource_id of role assignment
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.authorization.role_assignments.get resource_id="value"  raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.public_ip_addresses.get
                - kwargs:
                    name: "my_resource"
                    resource_id: "{scope}/providers/Microsoft.Authorization/roleAssignments/{role_assignment_name}"
                    raw: False

    """

    result = dict(comment=[], ret=None, result=True)

    uri_parameters = OrderedDict({"roleAssignments": "role_assignment_name"})
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/{resource_id}?api-version=2015-07-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"] = response_get["comment"]
        return result

    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )

    if raw:
        result["ret"] = response_get["ret"]

    else:
        result[
            "ret"
        ] = hub.tool.azure.authorization.role_assignments.convert_raw_role_assignments_to_present(
            resource=response_get["ret"],
            idem_resource_name=resource_id,
            role_assignment_name=uri_parameter_values["role_assignment_name"],
            resource_id=resource_id,
        )

    return result


async def list_(hub, ctx) -> Dict:
    """List of role assignment

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.authorization.role_assignments.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.authorization.role_assignments.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict({"roleAssignments": "role_assignment_name"})
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Authorization/roleAssignments?api-version=2015-07-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                result["ret"].append(
                    hub.tool.azure.authorization.role_assignments.convert_raw_role_assignments_to_present(
                        resource=resource,
                        idem_resource_name=resource_id,
                        role_assignment_name=uri_parameter_values[
                            "role_assignment_name"
                        ],
                        resource_id=resource_id,
                    )
                )
    return result



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_raw_virtual_network_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    virtual_network_name: str,
    resource_id: str,
    subscription_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        virtual_network_name: Azure Virtual Network resource name.
        resource_id: Azure Virtual Network resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "virtual_network_name": virtual_network_name,
        "location": resource["location"],
        "subscription_id": subscription_id,
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "bgpCommunities": "bgp_communities",
            "flowTimeoutInMinutes": "flow_timeout_in_minutes",
            "provisioningState": "provisioning_state",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties.get(parameter_raw)
        resource_translated["address_space"] = properties["addressSpace"][
            "addressPrefixes"
        ]
        if properties.get("subnets") is not None:
            existing_subnets_required_payload = convert_raw_to_present_vn_subnet(
                subnets=properties["subnets"]
            )
            resource_translated["subnets"] = existing_subnets_required_payload
    return resource_translated


def convert_present_to_raw_virtual_network(
    hub,
    subscription_id: str,
    address_space: List,
    location: str,
    bgp_communities: str = None,
    flow_timeout_in_minutes: int = None,
    subnets: List = None,
    tags: Dict = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id
        address_space: An array of IP address ranges that can be used by subnets of the virtual network.
        location: Resource location. Update this field will result in resource re-creation.
        bgp_communities: Bgp Communities sent over ExpressRoute with each route corresponding to a prefix in this VNET.
        flow_timeout_in_minutes: The FlowTimeout value (in minutes) for the Virtual Network
        subnets(list[dict], optional): List of Subnet in a virtual network resource.Each Subnet will have fields
          name(str, required), address_space(str, required) , security_group_id(str, optional) and service_endpoints(list, optional)
        tags: Resource tags.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "location": location,
        "properties": {"addressSpace": {"addressPrefixes": address_space}},
    }
    if tags is not None:
        payload["tags"] = tags
    if bgp_communities is not None:
        payload["properties"]["bgpCommunities"] = bgp_communities
    if flow_timeout_in_minutes is not None:
        payload["properties"]["flowTimeoutInMinutes"] = flow_timeout_in_minutes
    if subnets is not None:
        subnets_payload = convert_present_to_raw_vn_subnet(
            subscription_id=subscription_id, subnets=subnets
        )
        payload["properties"]["subnets"] = subnets_payload
    return payload


def update_virtual_network_payload(
    hub,
    subscription_id: str,
    existing_payload: Dict[str, Any],
    new_values: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id required to update network security group for VN subnet
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    is_updated = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        is_updated = True
    existing_properties = existing_payload["properties"]
    if (new_values.get("address_space") is not None) and (
        set(new_values["address_space"])
        != set(existing_properties["addressSpace"]["addressPrefixes"])
    ):
        new_payload["properties"]["addressSpace"]["addressPrefixes"] = new_values[
            "address_space"
        ]
        is_updated = True
    if (new_values.get("bgp_communities") is not None) and (
        new_values["bgp_communities"] != existing_properties.get("bgpCommunities")
    ):
        new_payload["properties"]["bgpCommunities"] = new_values.get("bgp_communities")
        is_updated = True
    if (new_values.get("flow_timeout_in_minutes") is not None) and (
        new_values["flow_timeout_in_minutes"]
        != existing_properties.get("flowTimeoutInMinutes")
    ):
        new_payload["properties"]["flowTimeoutInMinutes"] = new_values.get(
            "flow_timeout_in_minutes"
        )
        is_updated = True
    if new_values.get("subnets") is not None:
        existing_subnets = existing_properties.get("subnets")
        existing_subnets_required_payload = convert_raw_to_present_vn_subnet(
            subnets=existing_subnets
        )
        if diff_virtual_network_subnets(
            new_values.get("subnets"), existing_subnets_required_payload, "name"
        ):
            new_payload["properties"]["subnets"] = convert_present_to_raw_vn_subnet(
                subscription_id=subscription_id, subnets=new_values.get("subnets")
            )
            is_updated = True
    if is_updated:
        result["ret"] = new_payload
    return result


def convert_present_to_raw_vn_subnet(
    subscription_id: str, subnets: List[Dict[str, Any]]
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        subscription_id : subscription id to create network security group id
        subnets(list[dict], optional): List of Subnet in a virtual network resource.Each Subnet will have fields
        ((name(str), required), (address_space(str, required) and service_endpoints(list), optional))

    Returns:
        Subnets List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    raw_subnets: List = []
    for subnet in subnets:
        payload = {
            "name": subnet["name"],
            "properties": {"addressPrefix": subnet["address_prefix"]},
        }
        if subnet.get("security_group_id") is not None:
            payload["properties"]["networkSecurityGroup"] = {
                "id": subnet["security_group_id"]
            }
        if subnet.get("service_endpoints") is not None:
            services: List = []
            for endpoint in subnet.get("service_endpoints"):
                services.append({"service": endpoint})
            payload["properties"]["serviceEndpoints"] = services
        raw_subnets.append(payload)
    return raw_subnets


def convert_raw_to_present_vn_subnet(subnets: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        subnets(list[dict], optional): Resource List of Subnet in a virtual network resource.

    Returns:
         A subnet List that contains the parameters that match respective present function's input format.
    """
    present_subnets: List = []
    for subnet in subnets:
        new_subnet_payload = {
            "name": subnet.get("name"),
            "address_prefix": subnet["properties"]["addressPrefix"],
        }
        if subnet["properties"].get("networkSecurityGroup") is not None:
            new_subnet_payload["security_group_id"] = subnet["properties"][
                "networkSecurityGroup"
            ].get("id")
        if subnet["properties"].get("serviceEndpoints") is not None:
            services: List = []
            for endpoint in subnet["properties"]["serviceEndpoints"]:
                services.append(endpoint["service"])
            new_subnet_payload["service_endpoints"] = services
        present_subnets.append(new_subnet_payload)
    return present_subnets


def diff_virtual_network_subnets(
    new_values: List[Dict[str, any]],
    resource: List[Dict[str, any]],
    sorting_key: str,
):
    """
    Compares virtual network subnets to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        hub: The redistributed pop central hub.
        new_values: (List[Dict[str, any]]) Present value which will be given as input
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False
    """
    for entry in new_values:
        for key, value in entry.items():
            if isinstance(entry[key], list):
                entry[key].sort()
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        resource, key=lambda x: x[sorting_key]
    )



__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_group_name: str,
    resource_provider_namespace: str,
    parent_resource_path: str,
    resource_type: str,
    resource: str,
):
    """
    https://docs.microsoft.com/en-us/rest/api/authorization/permissions/list-for-resource
    """
    subscription_id = ctx.acct.subscription_id
    return await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourcegroups/{resource_group_name}/providers/{resource_provider_namespace}/{parent_resource_path}/{resource_type}/{resource}/providers/Microsoft.Authorization/permissions",
        params={"api-version": "2015-07-01"},
        success_codes=[200],
    )


async def list_(hub, ctx, resource_group_name: str):
    """
    https://docs.microsoft.com/en-us/rest/api/authorization/permissions/list-for-resource
    """
    subscription_id = ctx.acct.subscription_id
    return await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourcegroups/{resource_group_name}/providers/Microsoft.Authorization/permissions",
        params={"api-version": "2015-07-01"},
        success_codes=[200],
    )



"""Exec module for managing storage account."""
from collections import OrderedDict
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def list_(hub, ctx) -> Dict:
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "storageAccounts": "account_name",
        }
    )

    # 1. List all available storage accounts
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Storage/storageAccounts?api-version=2021-04-01",
        success_codes=[200],
    ):
        resource_storage_accounts = page_result.get("value")
        if resource_storage_accounts:
            # 2. Iterate over storage accounts to get associated containers
            for resource_storage_account in resource_storage_accounts:
                resource_id = resource_storage_account["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )

                account_dict = hub.tool.azure.storage_resource_provider.storage_accounts.convert_raw_storage_accounts_to_present(
                    resource=resource_storage_account,
                    idem_resource_name=resource_id,
                    subscription_id=subscription_id,
                    resource_id=resource_id,
                    **uri_parameter_values,
                )

                if account_dict:
                    account_name = account_dict.get("account_name")
                    resource_grp_name = account_dict.get("resource_group_name")
                    async for page_result_internal in hub.tool.azure.request.paginate(
                        ctx,
                        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_grp_name}"
                        f"/providers/Microsoft.Storage/storageAccounts/{account_name}/blobServices/default/containers?api-version=2022-09-01",
                        success_codes=[200],
                    ):
                        # 3. Get paginated list of storage_blobs
                        storage_blobs = page_result_internal.get("value")
                        if storage_blobs:
                            for blob_container in storage_blobs:
                                blob_dict = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
                                    subscription_id,
                                    resource_grp_name,
                                    account_name,
                                    blob_container,
                                )
                                result["ret"].append(blob_dict)
    return result


async def get(hub, ctx, resource_id: str) -> Dict[str, Any]:
    result = dict(comment=[], result=True, ret=None)
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscription_id",
            "resourceGroups": "resource_group_name",
            "storageAccounts": "account_name",
            "containerName": "container_name",
        }
    )

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{hub.exec.azure.URL}{resource_id}?api-version=2021-04-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"] = response_get["comment"]
        return result
    uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
        resource_id, uri_parameters
    )

    result[
        "ret"
    ] = hub.tool.azure.storage_resource_provider.storage_blob.convert_raw_storage_blob_to_present(
        uri_parameter_values.get("subscription_id"),
        uri_parameter_values.get("resource_group_name"),
        uri_parameter_values.get("account_name"),
        response_get["ret"],
    )

    return result
